local addon = LibStub("AceAddon-3.0"):NewAddon("CoolAid", "AceEvent-3.0")
local db
local anchor

local media = LibStub("LibSharedMedia-3.0")
local AceConfig = LibStub("AceConfig-3.0")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")

local GetSpellInfo = _G.GetSpellInfo
local pairs = _G.pairs
local strformat = _G.string.format
local strjoin = _G.string.join
local bitband = _G.bit.band
local outsider = _G.COMBATLOG_OBJECT_AFFILIATION_OUTSIDER

local cooldowns = {
	[(GetSpellInfo(1766))] = 8, -- Kick
	[(GetSpellInfo(47528))] = 10, -- Mind Freeze
	[(GetSpellInfo(6555))] = 10, -- Pummel
	[(GetSpellInfo(72))] = 12, -- Shield Bash
	[(GetSpellInfo(2139))] = 24, -- Counterspell
	[(GetSpellInfo(19647))] = 24, -- Spell Lock
	}

local defaults = {
	profile = {
		growth = true,
		texture = "default",
		scale = 1,
		width = 200,
		height = 10,
		font = "ABF",
		fontsize = 10,
		pos = { x = 0, y = 0, p = "CENTER", rp = "CENTER"},
		spells = {},
	},
}

local options = {
	type = "group",
	args = {
		toggle = {
			type = "execute",
			name = "Toggle anchor",
			desc = "Toggle the bar anchor frame",
			func = function()
					addon:ToggleAnchor()
				end,
			order = 1,
		},
		test = {
			type = "execute",
			name = "Test Bar",
			desc = "Display a test bar",
			func = function()
					addon:TestBar()
				end,
			order = 1,
		},
		bars = {
			type = "group",
			name = "Bar settings",
			desc = "Bar settings",
			order = 10,
			args = {
				growth = {
					type = "toggle",
					name = "Grow upwards",
					desc = "Toggle bars grow upwards/downwards from anchor",
					get = function () return db.growth end,
					set = function (info, v)
						db.growth = v
						addon:UpdateAnchor()
					end,
					order = 1,
				},
				height = {
					type = "range",
					name = "Height",
					desc = "Set the height of the bars",
					get = function() return db.height end,
					set = function(info, v)
							db.height = v
							addon:UpdateAnchor()
						end,
					min = 10,
					max = 100,
					step = 1,
					isPercent = false,
					order = 3,
				},
				scale = {
					type = "range",
					name = "Scale",
					desc = "Set the scale of the bars",
					get = function() return db.scale end,
					set = function(info, v)
							db.scale = v
							addon:UpdateAnchor()
						end,
					min = 0.1,
					max = 5,
					step = 0.01,
					isPercent = true,
					order = 4,
				},
				texture = {
					type = "select",
					dialogControl = 'LSM30_Statusbar',
					order = 6,
					name = "Texture",
					desc = "Set the texture for the timer bars",
					values = AceGUIWidgetLSMlists.statusbar,
					get = function() return db.texture end,
					set = function(i,v)
							db.texture = v
							addon:UpdateAnchor()
						end,
				},
				font = {
					type = "select",
					dialogControl = 'LSM30_Font',
					order = 6,
					name = "Font",
					desc = "Set the bar font",
					values = AceGUIWidgetLSMlists.font,
					get = function() return db.font end,
					set = function(i, v)
							db.font = v
							addon:UpdateAnchor()
						end,
					order = 7,
				},
				fontsize = {
					type = "range",
					name = "Font Size",
					desc = "Set the bar font size",
					min = 7,
					max = 48,
					step = 1,
					get = function(info) return db.fontsize end,
					set = function(info, value)
							db.fontsize = value
							addon:UpdateAnchor()
						end,
					order = 8,
				},
			},
		},
		spells = {
			type = 'group',
			name = 'Spells',
			desc = 'Toggle spell timer displays',
			order = 20,
			args = {},
		},
	},
}

for k in pairs(cooldowns) do
	defaults.profile.spells[k] = true
	options.args.spells.args[k] = {
		type = "toggle",
		name = k,
		get = function () return db.spells[k] end,
		set = function (i,v) db.spells[k] = v end,
	}
end

-- Credit to the BigWigs team (Rabbit, Ammo, et al) for the anchor code 
local function sortBars(a, b)
	return (a.remaining > b.remaining and db.growup) and true or false
end

local function rearrangeBars(anchor)
	local tmp = {}
	for bar in pairs(anchor.running) do
		tinsert(tmp, bar)
	end
	tsort(tmp, sortBars)
	local lastBar = nil
	for i, bar in ipairs(tmp) do
		bar:ClearAllPoints()
		if db.growup then
			bar:SetPoint("BOTTOMLEFT", lastBar or anchor, "TOPLEFT")
			bar:SetPoint("BOTTOMRIGHT", lastBar or anchor, "TOPRIGHT")
		else
			bar:SetPoint("TOPLEFT", lastBar or anchor, "BOTTOMLEFT")
			bar:SetPoint("TOPRIGHT", lastBar or anchor, "BOTTOMRIGHT")
		end
		lastBar = bar
	end
	wipe(tmp)
end

local function onDragHandleMouseDown(self)
	self:GetParent():StartSizing("BOTTOMRIGHT")
end

local function onDragHandleMouseUp(self, button)
	self:GetParent():StopMovingOrSizing()
end

local function onResize(self, width)
	db.width = width
	rearrangeBars(self)
end

local function onDragStart(self)
	self:StartMoving()
end

local function onDragStop(self)
	self:StopMovingOrSizing()
	local p, _, rp, x, y = self:GetPoint()
	db.pos.p = p
	db.pos.rp = rp
	db.pos.x = x
	db.pos.y = y
end

local function onControlEnter(self)
	GameTooltip:ClearLines()
	GameTooltip:SetOwner(self, "ANCHOR_TOPLEFT")
	GameTooltip:AddLine(self.tooltipHeader)
	GameTooltip:AddLine(self.tooltipText, 1, 1, 1, 1)
	GameTooltip:Show()
end

local function onControlLeave()
	GameTooltip:Hide()
end

local function getBar(text)
	local bar = nil
	for k in pairs(anchor.running) do
		if k.candyBarLabel:GetText() == text then
			bar = true
		end
	end
	return bar
end

local function stopBar(text)
	local bar = nil
	for k in pairs(anchor.running) do
		if (not text or k.candyBarLabel:GetText() == text) then
			k:Stop()
			bar = true
		end
	end
	if bar then rearrangeBars(anchor) end
	return bar
end

local function startBar(text, duration, icon)
	stopBar(text)
	local bar = candy:New(media:Fetch("statusbar", db.texture), db.width, db.height)
	bar:Set("anchor", anchor)
	anchor.running[bar] = true
	bar.candyBarBackground:SetVertexColor(unpack(db.color.bg))
	bar:SetColor(unpack(db.color.bar))
	bar.candyBarLabel:SetTextColor(unpack(db.color.text))
	bar.candyBarLabel:SetFont(media:Fetch("font", db.font), 10)
	bar.candyBarDuration:SetFont(media:Fetch("font", db.font), 10)
	bar:SetLabel(text)
	bar:SetDuration(duration)
	bar:SetTimeVisibility(true)
	bar:SetIcon(icon)
	bar:SetScale(db.scale)
	bar:Start()
	rearrangeBars(anchor)
end

local function runTest(anchor)
	local tmp = {}
	for i = 2, GetNumSpellTabs() do
		local name, texture, offset, numSpells = GetSpellTabInfo(i)
		if offset then
			for s = offset + 1, offset + numSpells do
				tinsert(tmp, (GetSpellName(s, BOOKTYPE_SPELL)))
			end
		end
	end
	local spell = tmp[random(1, #tmp)]
	local name, rank, icon = GetSpellInfo(spell)
	local duration = random(10, 30)
	startBar(name, nil, duration, icon)
	wipe(tmp)
end

local function toggleAnchor(anchor)
	if anchor:IsShown() then
		anchor:Hide()
	else
		anchor:Show()
	end
end

local function createAnchor(frameName, title)
	local display = CreateFrame("Frame", frameName, UIParent)
	display:EnableMouse(true)
	display:SetMovable(true)
	display:SetResizable(true)
	display:RegisterForDrag("LeftButton")
	display:SetWidth(db.width)
	display:SetHeight(20)
	display:SetMinResize(80, 20)
	display:SetMaxResize(1920, 20)
	display:ClearAllPoints()
	display:SetPoint(db.pos.p, UIParent, db.pos.rp, db.pos.x, db.pos.y)

	local bg = display:CreateTexture(nil, "PARENT")
	bg:SetAllPoints(display)
	bg:SetBlendMode("BLEND")
	bg:SetTexture(0, 0, 0, 0.3)

	local header = display:CreateFontString(nil, "OVERLAY")
	header:SetFontObject(GameFontNormal)
	header:SetText(title)
	header:SetAllPoints(display)
	header:SetJustifyH("CENTER")
	header:SetJustifyV("MIDDLE")

	local drag = CreateFrame("Frame", nil, display)
	drag:SetFrameLevel(display:GetFrameLevel() + 10)
	drag:SetWidth(16)
	drag:SetHeight(16)
	drag:SetPoint("BOTTOMRIGHT", display, -1, 1)
	drag:EnableMouse(true)
	drag:SetScript("OnMouseDown", onDragHandleMouseDown)
	drag:SetScript("OnMouseUp", onDragHandleMouseUp)
	drag:SetAlpha(0.5)

	local tex = drag:CreateTexture(nil, "BACKGROUND")
	tex:SetTexture("Interface\\AddOns\\CoolAid\\Textures\\draghandle")
	tex:SetWidth(16)
	tex:SetHeight(16)
	tex:SetBlendMode("ADD")
	tex:SetPoint("CENTER", drag)

	local test = CreateFrame("Button", nil, display)
	test:SetPoint("BOTTOMLEFT", display, "BOTTOMLEFT", 3, 3)
	test:SetHeight(14)
	test:SetWidth(14)
	test.tooltipHeader = "Test"
	test.tooltipText = "Creates a new test bar"
	test:SetScript("OnEnter", onControlEnter)
	test:SetScript("OnLeave", onControlLeave)
	test:SetScript("OnClick", function() runTest() end)
	test:SetNormalTexture("Interface\\AddOns\\CoolAid\\Textures\\test")

	local close = CreateFrame("Button", nil, display)
	close:SetPoint("BOTTOMLEFT", test, "BOTTOMRIGHT", 4, 0)
	close:SetHeight(14)
	close:SetWidth(14)
	close.tooltipHeader = "Hide"
	close.tooltipText = "Hides the anchor"
	close:SetScript("OnEnter", onControlEnter)
	close:SetScript("OnLeave", onControlLeave)
	close:SetScript("OnClick", function() toggleAnchor(anchor) end)
	close:SetNormalTexture("Interface\\AddOns\\CoolAid\\Textures\\close")

	display:SetScript("OnSizeChanged", onResize)
	display:SetScript("OnDragStart", onDragStart)
	display:SetScript("OnDragStop", onDragStop)
	display.running = {}
	display:Hide()
	return display
end

local function updateAnchor(anchor)
	for bar in pairs(anchor.running) do
		bar.candyBarBar:SetStatusBarTexture(media:Fetch("statusbar", db.texture))
		bar.candyBarBackground:SetTexture(media:Fetch("statusbar", db.texture))
		bar.candyBarBackground:SetVertexColor(unpack(db.color.bg))
		bar.candyBarBar:SetStatusBarColor(unpack(db.color.bar))
		bar.candyBarLabel:SetTextColor(unpack(db.color.text))
		bar.candyBarLabel:SetFont(media:Fetch("font", db.font), db.fontsize)
		bar.candyBarDuration:SetFont(media:Fetch("font", db.font), db.fontsize)
		bar:SetScale(db.scale)
	end	
	rearrangeBars(anchor)
end

function addon:UpdateProfile()
	db = self.db.profile
	updateAnchor(anchor)
end

function addon:RunTest()
	runTest(anchor)
end

function addon:ToggleAnchor()
	toggleAnchor(anchor)
end

function addon:UpdateAnchor()
	updateAnchor(anchor)
end

function addon:OnInitialize()
	addon.db = LibStub("AceDB-3.0"):New("CoolAidDB", defaults, "Default")
	db = addon.db.profile 

	AceConfig:RegisterOptionsTable("CoolAid", options)

	local optFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("CoolAid", "CoolAid")
	options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(addon.db)

	addon.db.RegisterCallback(self, "OnProfileChanged", "UpdateAnchor")
	addon.db.RegisterCallback(self, "OnProfileCopied", "UpdateAnchor")
	addon.db.RegisterCallback(self, "OnProfileReset", "UpdateAnchor")	

	anchor = createAnchor("CoolAidAnchor", "CoolAid")
end

function addon:OnEnable()
	addon:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
end

function addon:LibCandyBar_Stop(callback, bar)
	local a = bar:Get("anchor")
	if a == anchor and anchor.running and anchor.running[bar] then
		anchor.running[bar] = nil
	end
end

function addon:UpdateProfile()
	db = self.db.profile
	updateAnchor(anchor)
end

function addon:RunTest()
	runTest(anchor)
end

function addon:ToggleAnchor()
	toggleAnchor(anchor)
end

function addon:UpdateAnchor()
	updateAnchor(anchor)
end

function addon:COMBAT_LOG_EVENT_UNFILTERED(_,timestamp,event,srcGUID,srcName,srcFlags,dstGUID,dstName,dstFlags,spellID,spellName,_,extraID,extraName)
	if (event=="SPELL_CAST_SUCCESS" or event=="SPELL_INTERRUPT") and bitband(srcFlags,outsider)==0 and (cooldowns[spellName]) and db.spells[spellName] then
		local id = strjoin("-",srcGUID,spellID)
		local _,_,icon = GetSpellinfo(spellID)	
		local time = cooldowns[spellName]	
		local text
		if event=="SPELL_INTERRUPT" then 
			text = strformat("%s: %s (%s)", srcName, spellName, extraName)
		elseif dstName then
			text = strformat("%s: %s (%s)", srcName, spellName, dstName)
		else
			text = strformat("%s: %s", srcName, spellName)
		end
		startBar(text, time, icon)
	end
end

